@@ -1,47 +1,37 @@
-/******************************************************************************
-*   Copyright 2007-2009 by Aaron Seigo <aseigo@kde.org>                       *
-*   Copyright 2012 Sebastian Kügler <sebas@kde.org>                           *
-*                                                                             *
-*   This library is free software; you can redistribute it and/or             *
-*   modify it under the terms of the GNU Library General Public               *
-*   License as published by the Free Software Foundation; either              *
-*   version 2 of the License, or (at your option) any later version.          *
-*                                                                             *
-*   This library is distributed in the hope that it will be useful,           *
-*   but WITHOUT ANY WARRANTY; without even the implied warranty of            *
-*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU          *
-*   Library General Public License for more details.                          *
-*                                                                             *
-*   You should have received a copy of the GNU Library General Public License *
-*   along with this library; see the file COPYING.LIB.  If not, write to      *
-*   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,      *
-*   Boston, MA 02110-1301, USA.                                               *
-*******************************************************************************/
+/*
+    SPDX-FileCopyrightText: 2007-2009 Aaron Seigo <aseigo@kde.org>
+    SPDX-FileCopyrightText: 2012 Sebastian Kügler <sebas@kde.org>
+
+    SPDX-License-Identifier: LGPL-2.0-or-later
+*/
 
 #include "private/packagejobthread_p.h"
 
 #include "package.h"
 #include "config-package.h"
 
-#include <karchive.h>
-#include <kdesktopfile.h>
-#include <klocalizedstring.h>
-#include <ktar.h>
+#include <KArchive>
+#include <KLocalizedString>
+#include <KTar>
 #include <kzip.h>
-#include <KConfigGroup>
+
+
+#include <KCompressionDevice>
 
 #include <QDir>
 #include <QFile>
 #include <QIODevice>
 #include <QMimeType>
 #include <QMimeDatabase>
-#include <QRegExp>
+#include <QRegularExpression>
 #include <QJsonArray>
 #include <QDirIterator>
 #include <QJsonDocument>
 #include <qtemporarydir.h>
-
-#include <QDebug>
+#include <QProcess>
+#include <QUrl>
+#include <QStandardPaths>
+#include "kpackage_debug.h"
 
 namespace KPackage
 {
@@ -61,7 +51,8 @@ bool copyFolder(QString sourcePath, QString targetPath)
         target = QDir(targetPath);
     }
 
-    foreach (const QString &fileName, source.entryList(QDir::Files)) {
+    const auto lstEntries = source.entryList(QDir::Files);
+    for (const QString &fileName : lstEntries) {
         QString sourceFilePath = sourcePath + QDir::separator() + fileName;
         QString targetFilePath = targetPath + QDir::separator() + fileName;
 
@@ -69,8 +60,8 @@ bool copyFolder(QString sourcePath, QString targetPath)
             return false;
         }
     }
-
-    foreach (const QString &subFolderName, source.entryList(QDir::AllDirs | QDir::NoDotAndDotDot)) {
+    const auto lstEntries2 = source.entryList(QDir::AllDirs | QDir::NoDotAndDotDot);
+    for (const QString &subFolderName : lstEntries2) {
         QString sourceSubFolderPath = sourcePath + QDir::separator() + subFolderName;
         QString targetSubFolderPath = targetPath + QDir::separator() + subFolderName;
 
@@ -88,32 +79,6 @@ bool removeFolder(QString folderPath)
     return folder.removeRecursively();
 }
 
-
-QVariantMap convert(const QString &src)
-{
-    KDesktopFile df(src);
-    KConfigGroup c = df.desktopGroup();
-
-    static const QSet<QString> boolkeys = QSet<QString>()
-                                          << QStringLiteral("Hidden") << QStringLiteral("X-KDE-PluginInfo-EnabledByDefault");
-    static const QSet<QString> stringlistkeys = QSet<QString>()
-            << QStringLiteral("X-KDE-ServiceTypes") << QStringLiteral("X-KDE-PluginInfo-Depends")
-            << QStringLiteral("X-Plasma-Provides");
-
-    QVariantMap vm;
-    foreach (const QString &k, c.keyList()) {
-        if (boolkeys.contains(k)) {
-            vm[k] = c.readEntry(k, false);
-        } else if (stringlistkeys.contains(k)) {
-            vm[k] = c.readEntry(k, QStringList());
-        } else {
-            vm[k] = c.readEntry(k, QString());
-        }
-    }
-
-    return vm;
-}
-
 bool removeIndex(const QString& dir)
 {
     bool ok = true;
@@ -123,12 +88,12 @@ bool removeIndex(const QString& dir)
             QFile f(fileInfo.absoluteFilePath());
             if (!f.remove()) {
                 ok = false;
-                qWarning() << "Cannot remove kplugin index file: " << fileInfo.absoluteFilePath();
+                qCWarning(KPACKAGE_LOG) << "Cannot remove kplugin index file: " << fileInfo.absoluteFilePath();
             } else {
-                qDebug() << "Deleted index: " << fileInfo.absoluteFilePath();
+                qCDebug(KPACKAGE_LOG) << "Deleted index: " << fileInfo.absoluteFilePath();
             }
         } else {
-            qWarning() << "Cannot remove kplugin index file (not writable): " << fileInfo.absoluteFilePath();
+            qCWarning(KPACKAGE_LOG) << "Cannot remove kplugin index file (not writable): " << fileInfo.absoluteFilePath();
             ok = false;
         }
     }
@@ -140,7 +105,6 @@ Q_GLOBAL_STATIC_WITH_ARGS(QStringList, metaDataFiles, (QStringList(QLatin1String
 bool indexDirectory(const QString& dir, const QString& dest)
 {
     QVariantMap vm;
-    QVariantMap pluginsVm;
     vm[QStringLiteral("Version")] = QStringLiteral("1.0");
     vm[QStringLiteral("Timestamp")] = QDateTime::currentMSecsSinceEpoch();
 
@@ -164,13 +128,13 @@ bool indexDirectory(const QString& dir, const QString& dest)
 
     QString destfile = dest;
     if (!QDir::isAbsolutePath(dest)) {
-        destfile = dir + '/' + dest;
+        destfile = dir + QLatin1Char('/') + dest;
     }
 
     QDir().mkpath(QFileInfo(destfile).dir().absolutePath());
-    QFile file(destfile);
+    KCompressionDevice file(destfile, KCompressionDevice::BZip2);
     if (!file.open(QIODevice::WriteOnly)) {
-        qWarning() << "Failed to open " << destfile;
+        qCWarning(KPACKAGE_LOG) << "Failed to open " << destfile;
         return false;
     }
 
@@ -178,7 +142,7 @@ bool indexDirectory(const QString& dir, const QString& dest)
     jdoc.setArray(plugins);
 //     file.write(jdoc.toJson());
     file.write(jdoc.toBinaryData());
-    qWarning() << "Generated " << destfile << " (" << plugins.count() << " plugins)";
+    qCWarning(KPACKAGE_LOG) << "Generated " << destfile << " (" << plugins.count() << " plugins)";
 
     return true;
 }
@@ -211,6 +175,31 @@ bool PackageJobThread::install(const QString &src, const QString &dest)
     return ok;
 }
 
+static QString resolveHandler(const QString &scheme)
+{
+    QString candidatePath = QStringLiteral(CMAKE_INSTALL_FULL_LIBEXECDIR_KF5 "/kpackagehandlers/%1handler").arg(scheme);
+    if (qEnvironmentVariableIsSet("KPACKAGE_DEP_RESOLVERS_PATH")) {
+        candidatePath = QStringLiteral("%1/%2handler").arg(QString::fromUtf8(qgetenv("KPACKAGE_DEP_RESOLVERS_PATH")), scheme);
+    }
+    return QFile::exists(candidatePath) ? candidatePath : QString();
+}
+
+bool PackageJobThread::installDependency(const QUrl &destUrl)
+{
+    auto handler = resolveHandler(destUrl.scheme());
+    if (handler.isEmpty())
+        return false;
+
+    QProcess process;
+    process.setProgram(handler);
+    process.setArguments({ destUrl.toString() });
+    process.setProcessChannelMode(QProcess::ForwardedChannels);
+    process.start();
+    process.waitForFinished();
+
+    return process.exitCode() == 0;
+}
+
 bool PackageJobThread::installPackage(const QString &src, const QString &dest, OperationType operation)
 {
     QDir root(dest);
@@ -219,7 +208,7 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         if (!root.exists()) {
             d->errorMessage = i18n("Could not create package root directory: %1", dest);
             d->errorCode = Package::JobError::RootCreationError;
-            //qWarning() << "Could not create package root directory: " << dest;
+            //qCWarning(KPACKAGE_LOG) << "Could not create package root directory: " << dest;
             return false;
         }
     }
@@ -239,11 +228,11 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         // we have a directory, so let's just install what is in there
         path = src;
         // make sure we end in a slash!
-        if (path[path.size() - 1] != '/') {
-            path.append('/');
+        if (!path.endsWith(QLatin1Char('/'))) {
+            path.append(QLatin1Char('/'));
         }
     } else {
-        KArchive *archive = 0;
+        KArchive *archive = nullptr;
         QMimeDatabase db;
         QMimeType mimetype = db.mimeTypeForFile(src);
         if (mimetype.inherits(QStringLiteral("application/zip"))) {
@@ -255,14 +244,14 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
                    mimetype.inherits(QStringLiteral("application/x-lzma"))) {
             archive = new KTar(src);
         } else {
-            //qWarning() << "Could not open package file, unsupported archive format:" << src << mimetype.name();
+            //qCWarning(KPACKAGE_LOG) << "Could not open package file, unsupported archive format:" << src << mimetype.name();
             d->errorMessage = i18n("Could not open package file, unsupported archive format: %1 %2", src, mimetype.name());
             d->errorCode = Package::JobError::UnsupportedArchiveFormatError;
             return false;
         }
 
         if (!archive->open(QIODevice::ReadOnly)) {
-            //qWarning() << "Could not open package file:" << src;
+            //qCWarning(KPACKAGE_LOG) << "Could not open package file:" << src;
             delete archive;
             d->errorMessage = i18n("Could not open package file: %1", src);
             d->errorCode = Package::JobError::PackageOpenError;
@@ -270,7 +259,7 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         }
 
         archivedPackage = true;
-        path = tempdir.path() + '/';
+        path = tempdir.path() + QLatin1Char('/');
 
         d->installPath = path;
 
@@ -281,7 +270,7 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         if (entries.count() == 1) {
             const KArchiveEntry *entry = source->entry(entries[0]);
             if (entry->isDirectory()) {
-                path.append(entry->name()).append("/");
+                path = path + entry->name() + QLatin1Char('/');
             }
         }
 
@@ -293,12 +282,12 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
     KPluginMetaData meta;
     if (!entries.isEmpty()) {
         const QString metadataFilePath = entries.first().filePath();
-        if (metadataFilePath.endsWith(QLatin1String(".desktop")))
-            meta = KPluginMetaData(metadataFilePath);
-        else {
+        if (metadataFilePath.endsWith(QLatin1String(".desktop"))) {
+            meta = KPluginMetaData::fromDesktopFile(metadataFilePath, {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
+        } else {
             QFile f(metadataFilePath);
             if(!f.open(QIODevice::ReadOnly)){
-                qWarning() << "Couldn't open metadata file" << src << path;
+                qCWarning(KPACKAGE_LOG) << "Couldn't open metadata file" << src << path;
                 d->errorMessage = i18n("Could not open metadata file: %1", src);
                 d->errorCode = Package::JobError::MetadataFileMissingError;
                 return false;
@@ -309,7 +298,7 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
     }
 
     if (!meta.isValid()) {
-        qDebug() << "No metadata file in package" << src << path;
+        qCDebug(KPACKAGE_LOG) << "No metadata file in package" << src << path;
         d->errorMessage = i18n("No metadata file in package: %1", src);
         d->errorCode = Package::JobError::MetadataFileMissingError;
         return false;
@@ -317,9 +306,9 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
 
 
     QString pluginName = meta.pluginId();
-    qDebug() << "pluginname: " << meta.pluginId();
+    qCDebug(KPACKAGE_LOG) << "pluginname: " << meta.pluginId();
     if (pluginName.isEmpty()) {
-        //qWarning() << "Package plugin name not specified";
+        //qCWarning(KPACKAGE_LOG) << "Package plugin name not specified";
         d->errorMessage = i18n("Package plugin name not specified: %1", src);
         d->errorCode = Package::JobError::PluginNameMissingError;
         return false;
@@ -327,17 +316,17 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
 
     // Ensure that package names are safe so package uninstall can't inject
     // bad characters into the paths used for removal.
-    QRegExp validatePluginName("^[\\w-\\.]+$"); // Only allow letters, numbers, underscore and period.
-    if (!validatePluginName.exactMatch(pluginName)) {
-        //qDebug() << "Package plugin name " << pluginName << "contains invalid characters";
+    const QRegularExpression validatePluginName(QStringLiteral("^[\\w\\-\\.]+$")); // Only allow letters, numbers, underscore and period.
+    if (!validatePluginName.match(pluginName).hasMatch()) {
+        //qCDebug(KPACKAGE_LOG) << "Package plugin name " << pluginName << "contains invalid characters";
         d->errorMessage = i18n("Package plugin name %1 contains invalid characters", pluginName);
         d->errorCode = Package::JobError::PluginNameInvalidError;
         return false;
     }
 
     QString targetName = dest;
-    if (targetName[targetName.size() - 1] != '/') {
-        targetName.append('/');
+    if (targetName[targetName.size() - 1] != QLatin1Char('/')) {
+        targetName.append(QLatin1Char('/'));
     }
     targetName.append(pluginName);
 
@@ -346,12 +335,12 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
             KPluginMetaData oldMeta(targetName + QLatin1String("/metadata.desktop"));
 
             if (oldMeta.serviceTypes() != meta.serviceTypes()) {
-                d->errorMessage = i18n("The new package has a different type from the old version already installed.", meta.version(), meta.pluginId(), oldMeta.version());
+                d->errorMessage = i18n("The new package has a different type from the old version already installed.");
                 d->errorCode = Package::JobError::UpdatePackageTypeMismatchError;
             } else if (isVersionNewer(oldMeta.version(), meta.version())) {
                 const bool ok = uninstallPackage(targetName);
                 if (!ok) {
-                    d->errorMessage = i18n("Impossible to remove the old installation of %1 located at %2", pluginName, targetName);
+                    d->errorMessage = i18n("Impossible to remove the old installation of %1 located at %2. error: %3", pluginName, targetName, d->errorMessage);
                     d->errorCode = Package::JobError::OldVersionRemovalError;
                 }
             } else {
@@ -369,12 +358,23 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         }
     }
 
+    //install dependencies
+    const QStringList dependencies = KPluginMetaData::readStringList(meta.rawData(), QStringLiteral("X-KPackage-Dependencies"));
+    for(const QString &dep : dependencies) {
+        QUrl depUrl(dep);
+        if (!installDependency(depUrl)) {
+            d->errorMessage = i18n("Could not install dependency: '%1'", dep);
+            d->errorCode = Package::JobError::PackageCopyError;
+            return false;
+        }
+    }
+
     if (archivedPackage) {
         // it's in a temp dir, so just move it over.
         const bool ok = copyFolder(path, targetName);
         removeFolder(path);
         if (!ok) {
-            //qWarning() << "Could not move package to destination:" << targetName;
+            //qCWarning(KPACKAGE_LOG) << "Could not move package to destination:" << targetName;
             d->errorMessage = i18n("Could not move package to destination: %1", targetName);
             d->errorCode = Package::JobError::PackageMoveError;
             return false;
@@ -384,7 +384,7 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         // than move them
         const bool ok = copyFolder(path, targetName);
         if (!ok) {
-            //qWarning() << "Could not copy package to destination:" << targetName;
+            //qCWarning(KPACKAGE_LOG) << "Could not copy package to destination:" << targetName;
             d->errorMessage = i18n("Could not copy package to destination: %1", targetName);
             d->errorCode = Package::JobError::PackageCopyError;
             return false;
@@ -396,13 +396,8 @@ bool PackageJobThread::installPackage(const QString &src, const QString &dest, O
         tempdir.setAutoRemove(false);
     }
 
-
     indexDirectory(dest, QStringLiteral("kpluginindex.json"));
-
-
     d->installPath = targetName;
-
-    //qWarning() << "Not updating kbuildsycoca4, since that will go away. Do it yourself for now if needed.";
     return true;
 }
 
@@ -417,9 +412,9 @@ bool PackageJobThread::update(const QString &src, const QString &dest)
 bool PackageJobThread::uninstall(const QString &packagePath)
 {
     bool ok = uninstallPackage(packagePath);
-    //qDebug() << "emit installPathChanged " << d->installPath;
+    //qCDebug(KPACKAGE_LOG) << "emit installPathChanged " << d->installPath;
     emit installPathChanged(QString());
-    //qDebug() << "Thread: installFinished" << ok;
+    //qCDebug(KPACKAGE_LOG) << "Thread: installFinished" << ok;
     emit finished(ok, d->errorMessage);
     return ok;
 }
@@ -435,14 +430,14 @@ bool PackageJobThread::uninstallPackage(const QString &packagePath)
     QString root;
     {
         // FIXME: remove, pass in packageroot, type and pluginName separately?
-        QStringList ps = packagePath.split('/');
+        QStringList ps = packagePath.split(QLatin1Char('/'));
         int ix = ps.count() - 1;
-        if (packagePath.endsWith('/')) {
+        if (packagePath.endsWith(QLatin1Char('/'))) {
             ix = ps.count() - 2;
         }
         pkg = ps[ix];
         ps.pop_back();
-        root = ps.join('/');
+        root = ps.join(QLatin1Char('/'));
     }
 
     bool ok = removeFolder(packagePath);
@@ -459,7 +454,7 @@ bool PackageJobThread::uninstallPackage(const QString &packagePath)
 
 Package::JobError PackageJobThread::errorCode() const
 {
-    return (Package::JobError)d->errorCode;
+    return static_cast<Package::JobError>(d->errorCode);
 }
 
 } // namespace KPackage
